"""
files.py
"""

import os


def count_files(path: str, suffix: str) -> int:
    """
    find number of files in a certain path that have a certain suffix
    """
    count = 0
    for _root, _dir, files in os.walk(path):
        for file in files:
            if file.endswith(suffix):
                count += 1
    return count



"""
copy.py
"""

from collections.abc import Generator

from pydmt.api.builder import Builder, Node, SourceFile, TargetFile
from pydmt.utils.filesystem import copy_mkdir
from pydmt.utils.digest import sha1_file


class BuilderCopy(Builder):
    def __init__(self,
                 source: str,
                 target: str,
                 ):
        # super().__init__()
        self.source = source
        self.sources: list[Node] = [SourceFile(self.source)]
        self.target = target
        self.targets: list[Node] = [TargetFile(self.target)]

    def get_sources(self) -> list[Node]:
        return self.sources

    def get_targets(self) -> list[Node]:
        return self.targets

    def build(self):
        copy_mkdir(self.source, self.target)

    def yield_results(self) -> Generator[tuple[str, str], None, None]:
        yield sha1_file(self.target), self.target



"""
make.py
"""

import os
from pydmt.api.feature import Feature
from pydmt.builders.make import BuilderMake
from pydmt.core.pydmt import PyDMT


SOURCE_FILE = "Makefile"


class FeatureMake(Feature):
    def setup(self, pydmt: PyDMT) -> None:
        if not os.path.isfile(SOURCE_FILE):
            return
        pydmt.add_builder(BuilderMake(
            source=SOURCE_FILE,
            target="out/make.stamp",
        ))



"""
main.py
"""

import sys
import os
import os.path
import pathlib
import logging

import pylogconf.core
from pytconf import register_endpoint, register_main, config_arg_parse_and_launch

from pydmt.configs import ConfigSudo, ConfigFlow, ConfigOutput, ConfigLogging, ConfigVenv, \
    ConfigImport, ConfigReqs
from pydmt.core.pydmt import PyDMT
from pydmt.static import APP_NAME, VERSION_STR, DESCRIPTION, LOGGER_NAME
from pydmt.utils.subprocess import check_call

from pydmt.features.sphinx import FeatureSphinx
from pydmt.features.mako import FeatureMako
from pydmt.features.yaml import FeatureYaml
from pydmt.features.apt import FeatureApt
from pydmt.features.npm import FeatureNpm
from pydmt.features.venv import FeatureVenv
from pydmt.features.reqs import FeatureReqs
from pydmt.features.make import FeatureMake
from pydmt.features.gem import FeatureGem


def check_config_and_load():
    if not os.path.isfile(".pydmt.config"):
        print("pydmt: not a pydmt folder", file=sys.stderr)
        sys.exit(1)


def add_all_features(p):
    """
    This method should be automatic, order matters here which is not good, order should be derived and not set in stone
    """
    f = FeatureApt()
    f.setup(p)
    f = FeatureNpm()
    f.setup(p)
    f = FeatureGem()
    f.setup(p)
    f = FeatureVenv()
    f.setup(p)
    f = FeatureReqs()
    f.setup(p)
    f = FeatureMako()
    f.setup(p)
    f = FeatureSphinx()
    f.setup(p)
    f = FeatureMake()
    f.setup(p)
    f = FeatureYaml()
    f.setup(p)


def add_to_path():
    """
    This adds to PYTHONPATH various paths we need
    If you disable this then templates would not be able to find things like "config/python.py"
    """
    if ConfigImport.import_system:
        folder = "/etc/pydmt"
        if folder not in sys.path:
            sys.path.insert(0, folder)
    if ConfigImport.import_home:
        folder = os.path.join(str(pathlib.Path.home()), ".config/pydmt")
        if folder not in sys.path:
            sys.path.insert(0, folder)
    if ConfigImport.import_cwd:
        if "" not in sys.path:
            sys.path.insert(0, "")


@register_endpoint(
    description="Build the project",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
        ConfigVenv,
        ConfigReqs,
    ],
)
def build():
    check_config_and_load()

    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    add_all_features(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Build python virtual environment",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
        ConfigVenv,
        ConfigReqs,
    ],
)
def build_venv():
    check_config_and_load()

    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    f = FeatureVenv()
    f.setup(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Install python prerequisites",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
        ConfigVenv,
        ConfigReqs,
    ],
)
def build_reqs():
    check_config_and_load()

    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    f = FeatureReqs()
    f.setup(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Build tools",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
        ConfigVenv,
        ConfigReqs,
    ],
)
def build_tools():
    check_config_and_load()

    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    f = FeatureApt()
    f.setup(p)
    f = FeatureNpm()
    f.setup(p)
    f = FeatureGem()
    f.setup(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Clean all generated files by asking each builder to clean its generated files",
    configs=[
        ConfigLogging,
    ],
)
def clean() -> None:
    check_config_and_load()

    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    add_all_features(p)

    # shutil.rmtree(".pydmt")
    p.clean_all()


@register_endpoint(
    description="Clean all non git tracked files (including pydmt cache)",
    configs=[
        ConfigLogging,
    ],
)
def clean_hard() -> None:
    check_config_and_load()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    # shutil.rmtree(".pydmt")
    check_call(["git", "clean", "-qffxd"])


@register_main(
    app_name=APP_NAME,
    version=VERSION_STR,
    main_description=DESCRIPTION,
)
def main():
    pylogconf.core.setup()
    config_arg_parse_and_launch()


if __name__ == "__main__":
    main()



"""
apt.py
"""

import os

from pydmt.api.feature import Feature
from pydmt.builders.apt import BuilderApt
from pydmt.core.pydmt import PyDMT
from pydmt.helpers.attrs import get_packages, get_packages_remove


class FeatureApt(Feature):
    def __init__(
        self,
    ):
        self.packages = get_packages()
        self.packages_remove = get_packages_remove()

    def setup(self, pydmt: PyDMT) -> None:
        if self.packages is None:
            return
        if not os.path.isfile("config/deps.py"):
            return
        pydmt.add_builder(BuilderApt(
            source="config/deps.py",
            target="out/deps.stamp",
            packages=self.packages,
            packages_remove=self.packages_remove,
        ))



"""
filesystem.py
"""

import os
import shutil


def makedirs_for_file(filename: str):
    folder = os.path.dirname(filename)
    if folder != "":
        os.makedirs(folder, exist_ok=True)


def copy_mkdir(source: str, destination: str):
    makedirs_for_file(destination)
    shutil.copy(source, destination)


def unlink_files(files: list[str], only_if_exist: bool = True) -> None:
    for file in files:
        if only_if_exist:
            if os.path.isfile(file):
                os.unlink(file)
        else:
            os.unlink(file)


def files_under_folder(folder: str, suffix: str | None = None) -> list[str]:
    file_list = []
    for dir_path, _, filenames in os.walk(folder):
        for filename in filenames:
            if suffix is None or filename.endswith(suffix):
                file_list.append(os.path.join(dir_path, filename))
    return sorted(file_list)


def files_under_folders(folders: list[str]) -> list[str]:
    file_list = []
    for folder in folders:
        file_list.extend(files_under_folder(folder))
    return sorted(file_list)


def touch(filename: str) -> None:
    with open(filename, "w"):
        pass


def mkdir_touch(filename: str) -> None:
    makedirs_for_file(filename)
    touch(filename)


def remove_files_by_suffix(folder:str, suffix:str, exceptions: list[str]):
    for filename in os.listdir(folder):
        file= os.path.join(folder, filename)
        if (
            os.path.isfile(file)
            and filename.endswith(suffix)
            and filename not in exceptions
        ):
            os.remove(file)



"""
deb_python_package.py
"""

import pydmt.helpers.project

# deb section
deb_package = True
deb_section = "python"
deb_priority = "optional"
deb_architecture = "all"
deb_pkgname = pydmt.helpers.project.get_name()
# to which series to publish the package?
# https://wiki.ubuntu.com/Releases
deb_series = [
    # "eoan",  # 19.10
    "disco",  # 19.04
    "cosmic",  # 18.10
    "bionic",  # 18.04
    # "artful",  # 17.10 EOL
    # "zesty",  # 17.04 EOL
    # "yakkety",  # 16.10 EOL
    "xenial",  # 16.04
    # "wily",  # 15.10 EOL
    # "vivid",  # 15.04 EOL
    # "utopic", # 14.10 EOL
    "trusty",  # 14.04
    # "saucy", # 13.10 EOL
    # "raring", # 13.04 EOL
]
# these are dependencies for runtime
# deb_depends = f"${misc:Depends}, ${python3:Depends}, python3-mako"
# these are dependencies for buildtime at launchpad ppa
deb_builddepends = "python3, python3-setuptools, debhelper, dh-python"
deb_standards_version = "3.9.8"
# deb_standards_version="3.9.5"
deb_x_python_version = ">= 3.4"
deb_x_python3_version = ">= 3.4"
deb_urgency = "low"



"""
urls.py
"""

import pydmt.helpers.project
from pydmt.helpers.attrs import get_github_username


def get_website():
    github_username = get_github_username()
    if github_username is None:
        return None
    name = pydmt.helpers.project.get_name()
    if name is None:
        return None
    return f"https://{github_username}.github.io/{name}"


def get_website_source():
    github_username = get_github_username()
    if github_username is None:
        return None
    name = pydmt.helpers.project.get_name()
    if name is None:
        return None
    return f"https://github.com/{github_username}/{name}"


def get_website_git():
    github_username = get_github_username()
    if github_username is None:
        return None
    name = pydmt.helpers.project.get_name()
    if name is None:
        return None
    return f"git://github.com/{github_username}/{name}.git"


def get_website_ppa():
    launchpad_username = get_github_username()
    if launchpad_username is None:
        return None
    return f"https://launchpanet/~{launchpad_username}/+archive/ubuntu/ppa"



"""
npm.py
"""

import os.path

from pydmt.api.feature import Feature
from pydmt.builders.npm import Installer
from pydmt.core.pydmt import PyDMT


class FeatureNpm(Feature):
    def setup(self, pydmt: PyDMT) -> None:
        if not os.path.isfile("package.json"):
            return
        pydmt.add_builder(Installer(
            source="package.json",
            target="out/npm.stamp",
        ))



"""
pydmt.py
"""

import logging

import os

from pydmt.api.builder import Builder
from pydmt.core.cache import Cache
from pydmt.utils.filesystem import copy_mkdir
from pydmt.utils.digest import sha1_file
from pydmt.configs import ConfigFlow
from pydmt.static import LOGGER_NAME


class BuildProcessStats:
    def __init__(self):
        self.builder_fail = []
        self.builder_ok = []
        self.copy_missing = []
        self.copy_sha1 = []
        self.nop = []

    def add_copy_missing(self, filename: str, object_name: str):
        self.copy_missing.append((filename, object_name))

    def get_copy_missing(self):
        return len(self.copy_missing)

    def add_copy_sha1(self, filename: str, object_name: str):
        self.copy_sha1.append((filename, object_name))

    def add_nop(self, filename: str, object_name: str):
        self.nop.append((filename, object_name))

    def get_nop(self):
        return len(self.nop)

    def add_builder_ok(self, builder: Builder):
        self.builder_ok.append(builder)

    def add_builder_fail(self, builder: Builder, e: Exception):
        self.builder_fail.append((builder, e))

    def get_builder_ok(self):
        return len(self.builder_ok)

    def get_builder_fail(self):
        return len(self.builder_fail)

    def get_os_error_code(self) -> int:
        if len(self.builder_fail) > 0:
            return 1
        return 0


class PyDMT:
    def __init__(self):
        self.builders: list[Builder] = []
        self.target_to_builder: dict[str, Builder] = {}
        self.cache = Cache()

    def build_by_builder(self, builder: Builder, stats: BuildProcessStats):
        """ run one builder, return statistics about the run """
        logger = logging.getLogger(LOGGER_NAME)
        target_signature = builder.get_signature()
        logger.debug(f"examining [{builder.get_name()}]")
        if self.cache.list_sig_ok(target_signature):
            logger.debug(f"verifying [{builder.get_name()}]")
            file_bad = 0
            file_correct = 0
            file_missing = 0
            file_total = 0
            list_filename = self.cache.get_list_filename(target_signature)
            for object_name, signature in Cache.iterate_objects(list_filename):
                filename = self.cache.get_object_filename(signature)
                if os.path.isfile(object_name):
                    object_name_signature = sha1_file(object_name)
                    if object_name_signature != signature:
                        logger.debug(f"file [{object_name}] is incorrect. Getting from cache.")
                        copy_mkdir(filename, object_name)
                        stats.add_copy_sha1(filename, object_name)
                        file_bad += 1
                    else:
                        logger.debug(f"file [{object_name}] is up to date")
                        stats.add_nop(filename, object_name)
                        file_correct += 1
                else:
                    logger.debug(f"file [{object_name}] is missing. Getting from cache.")
                    copy_mkdir(filename, object_name)
                    stats.add_copy_missing(filename, object_name)
                    file_missing += 1
                file_total += 1
            if file_bad > 0 or file_missing > 0:
                logger.debug(f"Retrieved {file_total} files from cache")
                logger.debug(f"(bad/correct/missing = {file_bad}/{file_correct}/{file_missing}")
            else:
                logger.debug(f"ok [{builder.get_name()}]")
        else:
            logger.debug(f"running [{builder.get_name()}]")
            print(f"{builder.get_name()}: [{builder.get_targets_as_string()}]...", end="", flush=True)
            # this is one of the rare cases in which we really want to catch all exceptions.
            # pylint: disable=broad-except
            # noinspection PyBroadException
            try:
                builder.build()
            except Exception as e:
                print("FAIL")
                logger.exception("exception")
                stats.add_builder_fail(builder, e)
                return False
            print("OK")
            stats.add_builder_ok(builder)

            d = {}
            for signature, target in builder.yield_results():
                self.cache.save_object_by_signature(signature, target)
                d[target] = signature
            self.cache.save_list_by_signature(target_signature, d)
        return True

    def build_by_target(self, target: str, stats: BuildProcessStats) -> None:
        b = self.target_to_builder[target]
        self.build_by_builder(b, stats)

    def build_by_targets(self, targets: list[str], stats: BuildProcessStats) -> None:
        for target in targets:
            self.build_by_target(target, stats)

    def build_all(self) -> BuildProcessStats:
        """
        Build all the targets, very high level method
        TODO: order builders by depenencies and do this multi-core
        :return: statistics about the build
        """
        stats = BuildProcessStats()
        for builder in self.builders:
            res = self.build_by_builder(
                builder=builder,
                stats=stats,
            )
            if not res and ConfigFlow.stop_after_error:
                return stats
        return stats

    def clean_all(self) -> None:
        """
        Clean all targets
        """
        for builder in self.builders:
            for target in builder.get_targets():
                target.remove()

    def add_builder(self, b: Builder) -> None:
        self.builders.append(b)
        for target in b.get_targets():
            self.target_to_builder[target.get_name()] = b



"""
mako.py
"""

import os
from pydmt.api.feature import Feature
from pydmt.builders.mako import BuilderMako
from pydmt.builders.copy import BuilderCopy
from pydmt.core.pydmt import PyDMT
from pydmt.utils.filesystem import files_under_folder


class FeatureMako(Feature):
    def __init__(
        self,
        data=None,
        templates_folder: str = "templates",
        config_folder: str = "config",
        snipplet_folder: str = "snipplets",
    ):
        self.data = data
        self.templates_folder = templates_folder
        self.config_folder = config_folder
        self.snipplet_folder = snipplet_folder

    def setup(self, pydmt: PyDMT) -> None:
        if not os.path.isdir(self.templates_folder):
            return
        for root, _, filenames in os.walk(self.templates_folder):
            for filename in filenames:
                source = os.path.join(root, filename)
                target_base, ext = os.path.splitext(source)
                if ext == ".mako":
                    target = os.sep.join(target_base.split(os.sep)[1:])
                    builder = BuilderMako(
                        source=source,
                        target=target,
                        data=self.data,
                        config_files=files_under_folder(self.config_folder, suffix=".py"),
                        snipplet_files=files_under_folder(self.snipplet_folder, suffix=".mako"),
                    )
                    pydmt.add_builder(builder)
                else:
                    target = os.sep.join(source.split(os.sep)[1:])
                    builder_copy = BuilderCopy(
                        source=source,
                        target=target,
                    )
                    pydmt.add_builder(builder_copy)



"""
snipplets.py
"""


def get_google_analytics(project_google_analytics_tracking_id):
    return f"""<script type="text/javascript">
    (function(i,s,o,g,r,a,m){{i["GoogleAnalyticsObject"]=r;i[r]=i[r]||function(){{
    (i[r].q=i[r].q||[]).push(arguments)}},i[r].l=1*new Date();a=s.createElement(o),
    m=s.getElementsByTagName(o)[0];a.async=1;a.src=g;m.parentNode.insertBefore(a,m)
    }})(window,document,"script","https://www.google-analytics.com/analytics.js","ga");
    ga("create", "{project_google_analytics_tracking_id}", "auto");
    ga("send", "pageview");
    </script>"""


def get_paypal(project_paypal_donate_button_id):
    return f"""<form action="https://www.paypal.com/cgi-bin/webscr"
    method="post" target="_top">
    <input type="hidden" name="cmd" value="_s-xclick">
    <input type="hidden" name="hosted_button_id" value="{project_paypal_donate_button_id}">
    <input type="image" src="https://www.paypalobjects.com/en_US/IL/i/btn/btn_donateCC_LG.gif" name="submit"
    alt="PayPal - The safer, easier way to pay online!">
    <img alt="" border="0" src="https://www.paypalobjects.com/en_US/i/scr/pixel.gif" width="1" height="1">
    </form>"""



"""
venv.py
"""

import os
from pydmt.api.feature import Feature
from pydmt.builders.venv import BuilderVenv
from pydmt.core.pydmt import PyDMT


class FeatureVenv(Feature):
    def setup(self, pydmt: PyDMT) -> None:
        if "GITHUB_WORKFLOW" in os.environ:
            return
        if not os.path.isfile("config/python.py"):
            return
        if not os.path.isfile(".pydmt.venv"):
            return
        pydmt.add_builder(BuilderVenv(
            source="config/python.py",
            target="out/venv.stamp",
        ))



"""
github.py
"""

import importlib


def get_workflows_platforms():
    mod = importlib.import_module("config.github")
    if hasattr(mod, "workflows_platforms"):
        return getattr(mod, "workflows_platforms")
    return None



"""
tempdir.py
"""

import contextlib
import os
import shutil
import tempfile
from collections.abc import Callable


@contextlib.contextmanager
def cd(new_dir: str, cleanup: Callable):
    prev_dir = os.getcwd()
    os.chdir(os.path.expanduser(new_dir))
    try:
        yield
    finally:
        os.chdir(prev_dir)
        cleanup()


@contextlib.contextmanager
def tempdir():
    dir_path = tempfile.mkdtemp()

    def cleanup():
        shutil.rmtree(dir_path)
    with cd(dir_path, cleanup):
        yield dir_path



"""
mako.py
"""

import sys
import os
import os.path
from collections.abc import Generator

import mako
import mako.exceptions
import mako.lookup
import mako.template

from pydmt.api.builder import Builder, Node, SourceFile, TargetFile, SourceFolder
from pydmt.utils.filesystem import makedirs_for_file
from pydmt.utils.digest import sha1_file

FOLDER_CONFIG = "config"
FOLDER_SNIPPLETS = "snipplets"


class BuilderMako(Builder):
    # pylint: disable=too-many-positional-arguments
    def __init__(self,
                 source: str,
                 target: str,
                 data: dict[str, object] | None,
                 config_files: list[str],
                 snipplet_files: list[str],
                 ):
        # super().__init__()
        self.source = source
        self.target = target
        self.data = data
        self.config_files: list[str] = config_files
        self.snipplet_files: list[str] = snipplet_files
        self.sources: list[Node] = [SourceFile(self.source)]
        if os.path.isdir(FOLDER_CONFIG):
            self.sources.append(SourceFolder(FOLDER_CONFIG))
        if os.path.isdir(FOLDER_SNIPPLETS):
            self.sources.append(SourceFolder(FOLDER_SNIPPLETS))
        self.targets: list[Node] = [TargetFile(self.target)]

    def get_sources(self) -> list[Node]:
        return self.sources

    def get_targets(self) -> list[Node]:
        return self.targets

    def build(self):
        lookup = mako.lookup.TemplateLookup(
            directories=["."],
        )
        template = mako.template.Template(
            filename=self.source,
            lookup=lookup,
        )
        makedirs_for_file(self.target)
        if self.data is None:
            output = template.render()
        else:
            output = template.render(**self.data)
        with open(self.target, "w") as file_handle:
            file_handle.write(output)

    def yield_results(self) -> Generator[tuple[str, str], None, None]:
        yield sha1_file(self.target), self.target


def print_full_exception():
    print("printing full exception")
    traceback = mako.exceptions.RichTraceback()
    for (filename, line_number, function_name, line) in traceback.traceback:
        print(f"File {filename}, line {line_number}, in {function_name}")
        print(line)
    print(f"{traceback.error.__class__.__name__}: {traceback.error}")


def print_exception(e, input_file):
    found = False
    traceback = mako.exceptions.RichTraceback()
    for (filename, line_number, function_name, line) in traceback.traceback:
        if filename == input_file:
            print(f"{sys.argv[0]}: error {e} in {filename}, line {line_number} function {function_name}")
            print(f"{line}")
            found = True
    if not found:
        for (filename, line_number, function_name, line) in traceback.traceback:
            print(f"File {filename}, line {line_number}, in {function_name}")
            print(line)
        print(f"{traceback.error.__class__.__name__}: {traceback.error}")



"""
This module runs make
"""


import os
from pydmt.utils.filesystem import mkdir_touch
from pydmt.utils.subprocess import check_call_ve

from pydmt.api.one_source_one_target import OneSourceOneTarget


class BuilderMake(OneSourceOneTarget):
    def build(self) -> None:
        args = [
            "make",
        ]
        if "GITHUB_WORKFLOW" in os.environ:
            # TODO: Im assuming here that there is no previous MAKEFLAGS variable. Not good.
            # -j will cause endless parallel jobs to be run and that is why we add the number
            # of cores
            os.environ["MAKEFLAGS"] = f"-j{os.cpu_count()}"
        check_call_ve(args)
        mkdir_touch(self.target)



"""
This module installs npm packages
"""


from pydmt.utils.filesystem import unlink_files, mkdir_touch
from pydmt.utils.subprocess import check_call

from pydmt.api.one_source_one_target import OneSourceOneTarget


class Installer(OneSourceOneTarget):
    def build(self) -> None:
        unlink_files([self.target])
        args = [
            "npm",
            "install",
            "--quiet",
            "--silent",
        ]
        check_call(args)
        mkdir_touch(self.target)



"""
reqs.py
"""

import os
from pydmt.api.feature import Feature
from pydmt.builders.reqs import BuilderReqs
from pydmt.core.pydmt import PyDMT

SOURCE_FILE = "config/python.py"
TARGET_FILE = "out/reqs.stamp"


class FeatureReqs(Feature):
    def setup(self, pydmt: PyDMT) -> None:
        if not os.path.isfile(SOURCE_FILE):
            return
        builder = BuilderReqs(
            source=SOURCE_FILE,
            target=TARGET_FILE,
        )
        pydmt.add_builder(builder)



"""
python.y
"""

import os
import glob
import pprint
import importlib


def hlp_source_under(folder):
    """
    this function finds all the python packages under a folder and
    write the "packages" and "package_dir" entries for a python setup.py
    script
    """
    # walk the folder and find the __init__.py entries for packages.
    packages = []
    package_dir = {}
    for root, _, files in os.walk(folder):
        for file in files:
            if file != "__init__.py":
                continue
            full = os.path.dirname(os.path.join(root, file))
            relative = os.path.relpath(full, folder)
            packages.append(relative)
            package_dir[relative] = full
    # we use pprint because we want the order to always remain the same
    return f"packages={sorted(packages)},\npackage_dir={pprint.pformat(package_dir)}"


def hlp_files_under(destination_folder, pat):
    file_list = [x for x in glob.glob(pat) if os.path.isfile(x)]
    return f"(\'{destination_folder}\', {file_list})"


def make_hlp_project_keywords(d):
    def hlp_project_keywords():
        return f"{d.project_keywords.split()}"

    return hlp_project_keywords


def make_hlp_project_platforms(d):
    def hlp_project_platforms():
        return f"{d.project_platforms.split()}"

    return hlp_project_platforms


def make_hlp_project_classifiers(d):
    def hlp_project_classifiers():
        classifiers = d.project_classifiers.split("\n")
        classifiers = [x.strip()[1:-1] for x in classifiers]
        return f"{classifiers}"

    return hlp_project_classifiers


def make_hlp_wrap(level):
    def hlp_wrap(t):
        return t.replace("\n", "\n" + "\t" * level)

    return hlp_wrap


def collect_reqs(add_dev=False) -> list[str]:
    try:
        reqs = []
        mod = importlib.import_module("config.python")
        if hasattr(mod, "requires"):
            reqs += getattr(mod, "requires")
        if add_dev and hasattr(mod, "dev_requires"):
            reqs += getattr(mod, "dev_requires")
        return reqs
    except ModuleNotFoundError:
        return []


def collect_bootstrap_reqs() -> list[str]:
    try:
        mod = importlib.import_module("config.bootstrap")
        if hasattr(mod, "requires"):
            return getattr(mod, "requires")
        return []
    except ModuleNotFoundError:
        return []


def get_install_args():
    return [
        "python",
        "-m",
        "pip",
        "install",
    ]



"""
feature.py
"""

import abc

from pydmt.core.pydmt import PyDMT


class Feature(abc.ABC):
    """
    A Feature is a piece of code that adds builder to a pydmt system in a predefined
    way. This is a way to share common practices between projects.

    Examples of features can be:
    - all the rules about how to build a python module with documentation and release it.
    - all the rules about how to build a C project and release it.

    These are similar to facets or natures in development environments.
    """

    @abc.abstractmethod
    def setup(self, pydmt: PyDMT) -> None:
        """ build your feature here """



"""
php.py
"""


def to_php(x):
    if isinstance(x, str):
        return f"\'{x}\'"
    if isinstance(x, bool):
        if x:
            return "TRUE"
        return "FALSE"
    if x is None:
        return "null"
    raise ValueError("dont know how to translate type", type(x), x)



"""
git.py
"""

import subprocess


def get_git_last_tag() -> str:
    """ return the output of git describe --tag """
    return subprocess.check_output(
        ["git", "describe", "--abbrev=0", "--tags"],
        stderr=subprocess.DEVNULL,
    ).decode().rstrip()


def get_git_describe() -> str:
    """ return the output of git describe """
    return subprocess.check_output(
        ["git", "describe"],
        stderr=subprocess.DEVNULL,
    ).decode().rstrip()


def get_git_version() -> str:
    """ get version by git """
    return ".".join(get_git_describe().split("-"))


def count_files(pattern: str) -> int:
    """ Count the number of sources files of a certain pattern """
    return int(subprocess.check_output(f"git ls-files -- \"{pattern}\" | wc -l", shell=True).decode().rstrip())



"""
general.py
"""

import datetime
import os
import socket

current_folder = os.path.basename(os.getcwd())
current_year = datetime.datetime.now().year
homedir = os.path.expanduser("~")
hostname = socket.gethostname()



"""
project.py
"""

import importlib


def get_name():
    """ this gives you the name of the project, this must exist """
    mod = importlib.import_module("config.project")
    return getattr(mod, "name")



"""
Install requirements
"""

import os.path
from pydmt.api.one_source_one_target import OneSourceOneTarget
from pydmt.utils.subprocess import check_call
from pydmt.utils.filesystem import mkdir_touch
from pydmt.utils.python import collect_reqs, collect_bootstrap_reqs, get_install_args
from pydmt.configs import ConfigReqs

REQUIREMENTS = "requirements.txt"


class BuilderReqs(OneSourceOneTarget):
    """
    This is a review of how to build a python virtual environment:
    # create the virtualenv
    virtualenv [folder]
    # activate it
    source [folder]/bin/activate
    # install packages from file
    pip install -r requirements.txt
    # OR
    python -m pip install -r requirements.txt
    # OR if you want to install a list of packages:
    pip install -r [list of packages]
    # OR
    python -m pip install [list of packages]
    """
    def build(self) -> None:
        """
        If we check if there is a "requirements.txt" file with frozen requirements
        and install it if there is.

        If we do not have a frozen requirements we install from config/{python|bootstrap}.py
        Why do we do this in two stages? Whats wrong with doing this in one stage?
        Because importing python.py may fail because of prereqs that python.py
        needs. In this case the user specifies these prereqs in bootstrap.py
        """
        if os.path.isfile(REQUIREMENTS):
            args = get_install_args()
            args.extend(["-r", REQUIREMENTS])
            check_call(args)
            mkdir_touch(self.target)
            return
        args = get_install_args()
        packs = collect_bootstrap_reqs()
        if packs:
            args.extend(packs)
            check_call(args)
        args = get_install_args()
        packs = collect_reqs(add_dev=ConfigReqs.reqs_add_dev)
        if packs:
            args.extend(packs)
            check_call(args)
        mkdir_touch(self.target)



"""
gem.py
"""

import os.path

from pydmt.api.feature import Feature
from pydmt.builders.gem import Installer
from pydmt.core.pydmt import PyDMT


class FeatureGem(Feature):
    def setup(self, pydmt: PyDMT) -> None:
        if not os.path.isfile("Gemfile"):
            return
        pydmt.add_builder(Installer(
            source="Gemfile",
            target="out/gem.stamp",
        ))



"""
attrs.py
"""

import importlib
from pydmt.utils.importlib import module_exists


def get_github_username():
    if not module_exists("config"):
        return None
    if not module_exists("config.personal"):
        return None
    mod = importlib.import_module("config.personal")
    if not hasattr(mod, "github_username"):
        return None
    return getattr(mod, "github_username")


def get_launchpad_username():
    if not module_exists("config"):
        return None
    if not module_exists("config.personal"):
        return None
    mod = importlib.import_module("config.personal")
    if not hasattr(mod, "launchpad_username"):
        return None
    return getattr(mod, "launchpad_username")


def get_packages():
    if not module_exists("config"):
        return None
    if not module_exists("config.deps"):
        return None
    mod = importlib.import_module("config.deps")
    if not hasattr(mod, "packages"):
        return None
    return getattr(mod, "packages")


def get_packages_remove():
    if not module_exists("config"):
        return None
    if not module_exists("config.deps"):
        return None
    mod = importlib.import_module("config.deps")
    if not hasattr(mod, "packages_remove"):
        return None
    return getattr(mod, "packages_remove")



"""
signature.py
"""

import datetime
import time

import git


def get_project_year_started(repo=".") -> int:
    repo = git.Repo(repo)
    commits = list(repo.iter_commits())
    last_commit = commits[-1]
    t = time.gmtime(last_commit.authored_date)
    return t.tm_year


def get_copyright_years_long(repo="."):
    project_year_started = get_project_year_started(repo=repo)
    general_current_year = datetime.datetime.now().year
    return ", ".join(map(str, range(int(project_year_started), general_current_year + 1)))


def get_copyright_years_short(repo="."):
    project_year_started = get_project_year_started(repo=repo)
    general_current_year = datetime.datetime.now().year
    if general_current_year == project_year_started:
        return general_current_year
    return f"{project_year_started} - {general_current_year}"
